#!/usr/bin/env python3
#
#  conda.py
"""
Functions to aid building of conda packages.

.. versionadded:: 0.7.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import atexit
import difflib
import warnings
from contextlib import suppress
from datetime import datetime, timedelta
from itertools import chain
from typing import Iterable, List

# 3rd party
import apeye.slumber_url.exceptions
import platformdirs
from apeye.slumber_url import SlumberURL
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.stringlist import DelimitedList
from packaging.requirements import InvalidRequirement

# this package
from shippinglabel import __version__, normalize
from shippinglabel.requirements import ComparableRequirement, combine_requirements, read_requirements

__all__ = [
		"CONDA_API",
		"clear_cache",
		"get_channel_listing",
		"compile_requirements",
		"validate_requirements",
		"make_conda_description",
		"prepare_requirements",
		]

warnings.warn(
		"shippinglabel.conda is deprecated and will be removed in v2.0.0.\n"
		"Please use shippinglabel-conda instead.",
		DeprecationWarning,
		)

CONDA_API = SlumberURL("https://conda.anaconda.org", append_slash=False)
"""
Instance of :class:`apeye.slumber_url.SlumberURL` for accessing the Conda API.

.. versionadded:: 0.7.0
"""

atexit.register(CONDA_API.session.close)

cache_dir = PathPlus(
		platformdirs.user_cache_dir(
				appname="shippinglabel",
				appauthor="domdfcoding",
				version=__version__,
				)
		) / "conda_cache"


def clear_cache(*channel_name: str) -> None:
	r"""
	Clear the cached Conda channel listings.

	.. versionadded:: 0.7.0

	:param \*channel_name: The name(s) of the channels to clear the cache for.

	If no arguments are given the cache is cleared for all channels.

	.. latex:clearpage::
	"""

	filenames: Iterable[PathPlus]

	if channel_name:
		filenames = (cache_dir / f"{channel}.json" for channel in channel_name)
	else:
		filenames = cache_dir.glob("*.json")

	for filename in filenames:
		filename.unlink(missing_ok=True)


def get_channel_listing(channel_name: str) -> List[str]:
	"""
	Obtain the list of packages in the given Conda channel, either from the cache or from the Conda API.

	Responses are cached for 48 hours. The cache can be cleared manually with :func:`~.clear_cache`.

	.. versionadded:: 0.7.0

	:param channel_name:

	:raises ValueError: if the channel can't be found (*New in version 0.15.0*).
	"""

	cache_dir.maybe_make(parents=True)

	filename = cache_dir / f"{channel_name}.json"

	if filename.is_file():
		data = filename.load_json()
		if datetime.fromtimestamp(data["expires"]) > datetime.now():
			return data["packages"]

	conda_packages = set()

	try:
		for package in (CONDA_API / channel_name / "noarch" / "repodata.json").get()["packages"].values():
			conda_packages.add(package["name"])
	except apeye.slumber_url.exceptions.HttpNotFoundError:
		raise ValueError(f"Conda channel {channel_name!r} not found.")

	with suppress(apeye.slumber_url.exceptions.HttpNotFoundError):
		# TODO: other architectures
		for package in (CONDA_API / channel_name / "linux-64" / "repodata.json").get()["packages"].values():
			conda_packages.add(package["name"])

	data = {"expires": (datetime.now() + timedelta(hours=48)).timestamp(), "packages": sorted(conda_packages)}

	filename.dump_json(data, indent=2)

	return data["packages"]


def prepare_requirements(requirements: Iterable[ComparableRequirement]) -> Iterable[ComparableRequirement]:
	"""
	Prepare a list of requirements for use with conda.

	This entails removing any extras and markers from the requirements, and skipping any requirements with URLs,
	as conda does not support these.

	.. versionadded:: 0.13.0

	:param requirements:
	"""

	for requirement in sorted(combine_requirements(requirements)):
		if requirement.url:  # pragma: no cover
			continue

		# TODO: add the extra requirements
		if requirement.extras:
			requirement.extras = set()
		if requirement.marker:
			requirement.marker = None

		yield requirement


def compile_requirements(
		repo_dir: PathPlus,
		extras: Iterable[str] = (),
		) -> List[ComparableRequirement]:
	"""
	Compile a list of requirements for the package from the :file:`requirements.txt` file,
	and any extra dependencies.

	.. versionadded:: 0.7.0

	:param repo_dir:
	:param extras: A list of additional, optional requirements.
		These would be specified in "extras_require" for setuptools.
	"""  # noqa: D400

	requirements = chain(
			read_requirements(repo_dir / "requirements.txt")[0],
			[ComparableRequirement(r) for r in extras],
			)

	return list(prepare_requirements(requirements))


def validate_requirements(
		requirements: Iterable[ComparableRequirement],
		conda_channels: Iterable[str],
		) -> List[ComparableRequirement]:
	"""
	Ensure that all requirements are available from the given Conda channels,
	and normalize the names to those in the Conda channel.

	.. versionadded:: 0.7.0

	:param requirements:
	:param conda_channels:
	"""  # noqa: D400

	validated_requirements = []

	conda_packages = set()
	channels = DelimitedList(conda_channels)

	for channel in channels:
		for package in get_channel_listing(channel):
			conda_packages.add(package)

	for requirement in requirements:

		# Check alias_mapping first
		if requirement.name in alias_mapping:
			requirement.name = alias_mapping[requirement.name]
			validated_requirements.append(requirement)
			continue
		elif requirement.name in conda_packages:
			validated_requirements.append(requirement)
			continue

		matches = difflib.get_close_matches(requirement.name, conda_packages)
		for match in matches:
			if normalize(match) == normalize(requirement.name):
				requirement.name = match
				validated_requirements.append(requirement)
				break
		else:
			raise InvalidRequirement(
					f"Cannot satisfy the requirement {requirement.name!r} "
					f"from any of the channels: '{channels:', '}'."
					)

	return validated_requirements

	#  entry_points:
	#    - {{ import_name }} = {{ import_name }}:main
	#  skip_compile_pyc:
	#    - "*/templates/*.py"          # These should not (and cannot) be compiled


alias_mapping = {"ruamel-yaml": "ruamel.yaml"}
"""
Mapping of normalised names to names on Conda, if they differ for some reason.

.. versionadded:: 0.7.0
"""

# Really just due to https://github.com/conda-forge/ruamel.yaml-feedstock/issues/7
# TODO: might this be solved by the workaround for poetry?


def make_conda_description(summary: str, conda_channels: Iterable[str] = ()) -> str:
	"""
	Create a description for the Conda package from its summary and a list of channels required to install it.

	The description will look like::

		This is my fancy Conda package. Hope you like it 😉.


		Before installing please ensure you have added the following channels: conda-forge, bioconda

	if called as follows:

	.. code-block:: python

		make_conda_description(
			"This is my fancy Conda package. Hope you like it 😉.",
			["conda-forge", "bioconda"],
		)

	.. versionadded:: 0.8.0

	:param summary:
	:param conda_channels:
	"""

	conda_description = summary
	conda_channels = tuple(conda_channels)

	if conda_channels:
		conda_description += "\n\n\n"
		conda_description += "Before installing please ensure you have added the following channels: "
		conda_description += ", ".join(conda_channels)
		conda_description += '\n'

	return conda_description


if __name__ == "__main__":
	clear_cache()



"""
shippinglabel._vendor is for vendoring dependencies to work around boostrapping issues upstream.
Files inside shippinglabel._vendor should be considered immutable and only updated to new versions from upstream.
"""

import importlib

try:
	trove_classifiers = importlib.import_module("trove_classifiers", "trove_classifiers")
except ImportError:
	from . import trove_classifiers as trove_classifiers  # noqa: F401



#!/usr/bin/env python3
#
#  __init__.py
"""
Utilities for handling packages.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import re
import warnings
from typing import Dict, Iterable, List, Optional

# 3rd party
import dist_meta.distributions
from dist_meta.metadata_mapping import MetadataMapping
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from packaging.version import InvalidVersion, Version

__all__ = [
		"no_dev_versions",
		"no_pre_versions",
		"normalize",
		"normalize_keep_dot",
		"read_pyvenv",
		"get_project_links",
		]

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2020 Dominic Davis-Foster"
__license__: str = "MIT License"
__version__: str = "1.7.2"
__email__: str = "dominic@davis-foster.co.uk"


def no_dev_versions(versions: Iterable[str]) -> List[str]:
	"""
	Returns the subset of ``versions`` which does not end with ``-dev``.

	:param versions:
	"""

	return [v for v in versions if not v.endswith("-dev")]


def no_pre_versions(versions: Iterable[str]) -> List[str]:
	"""
	Returns the subset of ``versions`` which are not prereleases (alpha, beta, dev, rc etc.).

	.. versionadded:: 0.15.0

	:param versions:
	"""

	output = []

	for v in versions:
		try:
			if not Version(v).is_prerelease:
				output.append(v)
		except InvalidVersion:
			output.append(v)

	return output


_normalize_pattern = re.compile(r"[-_.]+")
_normalize_keep_dot_pattern = re.compile(r"[-_]+")


def normalize(name: str) -> str:
	"""
	Normalize the given name for PyPI et al.

	From :pep:`503` (public domain).

	:param name: The project name.
	"""

	return _normalize_pattern.sub('-', name).lower()


def normalize_keep_dot(name: str) -> str:
	"""
	Normalize the given name for PyPI et al., but keep dots in namespace packages.

	.. versionadded:: 0.2.1

	:param name: The project name.
	"""

	return _normalize_keep_dot_pattern.sub('-', name).lower()


def read_pyvenv(venv_dir: PathLike) -> Dict[str, str]:
	"""
	Reads the ``pyvenv.cfg`` for the given virtualenv, and returns a ``key: value`` mapping of its contents.

	.. versionadded:: 0.9.0

	:param venv_dir:
	"""

	pyvenv_config: Dict[str, str] = {}

	for line in (PathPlus(venv_dir) / "pyvenv.cfg").read_lines():
		if line:
			key, value, *_ = line.split(" = ")
			pyvenv_config[key] = value

	return pyvenv_config


class ProjectLinks(MetadataMapping):
	pass


def get_project_links(
		project_name: str,
		path: Optional[Iterable[PathLike]] = None,
		) -> MetadataMapping:
	"""
	Returns the web links for the given project.

	The exact keys vary, but common keys include "Documentation" and "Issue Tracker".

	.. versionadded:: 0.12.0

	:param project_name:
	:param path: The directories entries to search for distributions in.
		This can be used to search in a different (virtual) environment.
	:default path: :py:data:`sys.path`

	:rtype:

	.. versionchanged:: 1.0.0  Now returns a :class:`dist_meta.metadata_mapping.MetadataMapping` object.

	.. versionchanged:: 1.2.0

		The :core-meta:`Home-Page` field from Python core metadata is included under the ``Homepage`` key, if present.
		This matches the output parsed from PyPI for packages which are not installed.

	.. versionchanged:: 1.7.0  Added the ``path`` argument.
	"""

	# this package
	from shippinglabel.pypi import get_metadata

	warnings.warn(
			"shippinglabel.get_project_links is deprecated and will be removed in v2.0.0.\n"
			"Please import from the shippinglabel_pypi package instead.",
			DeprecationWarning,
			)

	# Try a local package first

	urls = ProjectLinks()

	try:
		dist = dist_meta.distributions.get_distribution(project_name, path=path)
		meta = dist.get_metadata()
		raw_urls = meta.get_all("Project-URL", default=())

		for url in raw_urls:
			label, url, *_ = map(str.strip, url.split(','))
			urls[label] = url

		if "Home-Page" in meta:
			urls["Homepage"] = meta["Home-Page"]

	except dist_meta.distributions.DistributionNotFoundError:
		# Fall back to PyPI

		metadata = get_metadata(project_name)["info"]

		if "project_urls" in metadata and metadata["project_urls"]:
			for label, url in metadata["project_urls"].items():
				urls[label] = url

	return urls



#!/usr/bin/env python3
#
#  checksum.py
"""
Utilities for creating and checking file sha256 checksums.

.. versionadded:: 0.6.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import io
from base64 import urlsafe_b64encode
from hashlib import md5, sha256
from typing import IO, TYPE_CHECKING, Callable, Optional, Union, cast

# 3rd party
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike

if TYPE_CHECKING:
	# stdlib
	from hashlib import _Hash
else:
	try:
		# stdlib
		from _hashlib import HASH as _Hash
	except ImportError:  # pragma: no cover
		try:
			# stdlib
			from _hashlib import Hash as _Hash
		except ImportError:
			pass

__all__ = ["get_sha256_hash", "check_sha256_hash", "get_md5_hash", "get_record_entry"]

_HashType = type(sha256())


def _get_hash(fp: IO[bytes], hash_type: Callable, blocksize: int = 1 << 20) -> "_Hash":
	"""
	Returns the SHA256 hash object for the given file object.

	:param fp:
	:param hash_type:
	:param blocksize: The blocksize to read the file with.

	:rtype: :mod:`hashlib.sha256() <hashlib>`
	"""

	hash_obj = hash_type()

	fb = fp.read(blocksize)
	while len(fb):  # pylint: disable=len-as-condition
		hash_obj.update(fb)
		fb = fp.read(blocksize)

	return hash_obj


def get_sha256_hash(
		filename: Union[PathLike, IO[bytes]],
		blocksize: int = 1 << 20,
		) -> "_Hash":
	"""
	Returns the SHA256 hash object for the given file.

	.. versionadded:: 0.6.0

	:param filename:
	:param blocksize: The blocksize to read the file with.

	:rtype: :mod:`hashlib.sha256() <hashlib>`

	.. versionchanged:: 0.16.0  Added support for already open file objects.
	"""

	if isinstance(filename, io.BufferedIOBase):
		return _get_hash(filename, sha256, blocksize)

	with open(cast(PathLike, filename), "rb") as f:
		return _get_hash(f, sha256, blocksize)


def get_md5_hash(
		filename: Union[PathLike, IO[bytes]],
		blocksize: int = 1 << 20,
		) -> "_Hash":
	"""
	Returns the md5 hash object for the given file.

	.. versionadded:: 0.15.0

	:param filename:
	:param blocksize: The blocksize to read the file with.

	:rtype: :mod:`hashlib.md5() <hashlib>`

	.. versionchanged:: 0.16.0  Added support for already open file objects.
	"""

	if isinstance(filename, io.BufferedIOBase):
		return _get_hash(filename, md5, blocksize)

	with open(cast(PathLike, filename), "rb") as f:
		return _get_hash(f, md5, blocksize)


def check_sha256_hash(
		filename: Union[PathLike, IO[bytes]],
		hash: Union["_Hash", str],  # noqa: A002  # pylint: disable=redefined-builtin
		blocksize: int = 1 << 20,
		) -> bool:
	r"""
	Returns whether the SHA256 hash for the file matches ``hash``.

	.. versionadded:: 0.6.0

	:param filename:
	:param hash: If a string, the hexdigest of the hash.
	:type hash: :py:obj:`~typing.Union`\[:mod:`hashlib.sha256() <hashlib>`, :class:`str`\]
	:param blocksize: The blocksize to read the file with.

	:rtype:

	.. versionchanged:: 0.16.0  Added support for already open file objects.
	.. latex:clearpage::
	"""

	if isinstance(hash, _HashType):
		hash = hash.hexdigest()  # noqa: A001  # pylint: disable=redefined-builtin

	return hash == get_sha256_hash(filename, blocksize).hexdigest()


def get_record_entry(filename: PathLike, blocksize: int = 1 << 20, relative_to: Optional[PathLike] = None) -> str:
	"""
	Constructs a :pep:`376` ``RECORD`` entry for the file.

	.. versionadded:: 0.6.0

	:param filename:
	:param blocksize: The blocksize to read the file with.
	:param relative_to:
	"""

	sha256_hash = get_sha256_hash(filename, blocksize).digest()
	digest = "sha256=" + urlsafe_b64encode(sha256_hash).decode("latin1").rstrip('=')

	filename = PathPlus(filename)

	length = filename.stat().st_size

	if relative_to:
		filename = filename.relative_to(relative_to)

	return ','.join([filename.as_posix(), digest, str(length)])
